/*
 * DB Database Design Form
 * Copyright (C) 1998-2001 by Tom Dyas (tdyas@users.sourceforge.net)
 * Copyright (C) 2002-2003 by Chalain Marc (maarc-chalain@voila.fr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * The design form is activated from two different contexts in the
 * program. The first context is when creating a new data source. In
 * this case. the following hold:
 *
 *   CurrentSource == NULL
 *   DesignViewDriver points at the data source driver to use
 *   DesignViewInitialSchema:
 *     NULL for blank schema 
 *     points at schema to fill the form with (freed when form closes)
 *   DesignViewName points at the name to use (freed when form closes)
 *
 * The second context is when an existing data source's schema is
 * being viewed or modified. In this case, the following hold:
 *
 *   CurrentSource points at the open data source
 *   DesignViewDriver == CurrentSource->driver
 *   DesignViewInitialSchema == NULL (ignored by code)
 *   DesignViewName == NULL (ignored by code)
 *
 * The difference between the modes can be determined based on the
 * state of the CurrentSource variable.
 */
#define ALLOW_ACCESS_TO_INTERNALS_OF_LISTS

#include <PalmOS.h>
#include <FloatMgr.h>	// Include Floating Point headers
#ifdef CONFIG_HANDERA
#include <Vga.h>
#endif

#include "db.h"
#include "enum.h"
#include "util.h"
#ifdef CONFIG_FLOAT
#include "double.h"
#endif
#ifdef CONFIG_SCRIPTS
#include "calcvalue.h"
#endif
#include "prefs.h"


#define GetObjectPtr(f,i) FrmGetObjectPtr((f),FrmGetObjectIndex((f),(i)))

#define col_popup   0
#define col_oldname 1
#define col_name    2
#define col_type    3

#define SCRIPTS_ARE_COMPATIBLE 0xFFFF


static UInt16 FirstField;
static MemHandle SchemaFieldsHandle = 0;
static Char *FieldString;
//this variable are used in script.c it's not a static.
SchemaDesignField * SchemaFields = 0;
UInt16 NumFields = 0;
static UInt32 NextUniqueID = 0;
static char ** popup_strings = 0;
static UInt16 popup_strings_count = 0;
static Boolean ForceRebuild = false;


static struct types_t {
    DataSourceFieldType type;
    UInt16 nameID;
    Int16 index;
} types[] = {
    { FIELD_TYPE_STRING,  stringID_String,  -1 },
    { FIELD_TYPE_NOTE,    stringID_Note,    -1 },
    { FIELD_TYPE_BOOLEAN, stringID_Boolean, -1 },
    { FIELD_TYPE_INTEGER, stringID_Integer, -1 },
#ifdef CONFIG_FLOAT
    { FIELD_TYPE_FLOAT,   stringID_Float,   -1 },
#endif
#ifdef CONFIG_SCRIPTS
    { FIELD_TYPE_CALCULATED, stringID_Calculated, -1 },
#endif
    { FIELD_TYPE_DATE,    stringID_Date,    -1 },
    { FIELD_TYPE_TIME,    stringID_Time,    -1 },
    { FIELD_TYPE_LIST,    stringID_List,    -1 },
    { FIELD_TYPE_LINK,    stringID_Link,    -1 },
    { FIELD_TYPE_LINKED,  stringID_Linked,  -1 },
    { 0 , 0, -1},
};

/* Global variables to allow other portions of the DB code to tell
 * this form if a blank schema should be used or if there is some
 * initial schema that should be used to fill the display.
 */

DataSourceSchema *DesignViewInitialSchema = 0;
DataSourceDriver *DesignViewDriver = 0;
char *DesignViewName = 0;

static void DrawPopupIndicator(void *, Int16, Int16, RectangleType *) SECT_UI3;
static void UpdateScrollers(void) SECT_UI3;
static Err LoadData(void *, Int16, Int16, Boolean, MemHandle *, Int16 *,
                    Int16 *, FieldPtr) SECT_UI3;
static Boolean SaveData(void *, Int16, Int16) SECT_UI3;
static void LoadTable(TablePtr) SECT_UI3;
static void SetColumnWidths(TablePtr) SECT_UI3;
static void InitTable(TablePtr) SECT_UI3;
static void Scroll(WinDirectionType) SECT_UI3;
static UInt32 * CopyLocalSchema(DataSourceSchema * dst,
                     SchemaDesignField * src, UInt16 numFields) SECT_UI3;
static Boolean UpdateNewDesign(void) SECT_UI3;
static Boolean UpdateExistingDesign(void) SECT_UI3;
static void InsertField(UInt16) SECT_UI3;
static void DeleteField(UInt16) SECT_UI3;
static void HandleContextMenu(FormType *, TableType *, Int16) SECT_UI3;
static void ShowFieldProperties(UInt16 whichField) SECT_UI3;
static Boolean DesignViewUpdateForm(FormPtr form, UInt16 type) SECT_UI3;
static void ClearDesignSchemaField( Int16 i) SECT_UI3;
static UInt16 CheckScripts() SECT_UI3;
static UInt16 GetNumberOfRows(TablePtr table, UInt16 height) SECT_UI3;

static void
DrawPopupIndicator(void * table, Int16 row, Int16 col, RectangleType *b)
{
    Coord x, y, width;
 
    width = 8;
    x = b->topLeft.x + 1;
    y = b->topLeft.y + (TblGetRowHeight(table, row) - (width / 2)) / 2;

    while (width > 0) {
        WinDrawLine(x, y, x + width - 1, y);
        x += 1;
        y += 1;
        width -= 2;
    }

} 

static void
UpdateScrollers(void)
{
    UInt16 upIndex, downIndex, fieldNum;
    Boolean scrollUp, scrollDown;
    Int16 row;
    FormPtr form;
    TablePtr table;

    scrollUp = FirstField != 0 && NumFields > 0;

    form = FrmGetActiveForm();
    table = GetObjectPtr(form, ctlID_DesignView_Table);
    row = TblGetLastUsableRow(table);
    if (row != -1 && row != 0) {
        fieldNum = TblGetRowID(table, row);
        scrollDown = fieldNum < NumFields - 1;
    } else {
        scrollDown = false;
    }

    upIndex = FrmGetObjectIndex(form, ctlID_DesignView_UpButton);
    downIndex = FrmGetObjectIndex(form, ctlID_DesignView_DownButton);
    FrmUpdateScrollers(form, upIndex, downIndex, scrollUp, scrollDown);
}

static Err
LoadData(void * table, Int16 row, Int16 column,
         Boolean editing, MemHandle * textH, Int16 * textOffset,
         Int16 * textAllocSize, FieldPtr fld)
{
    UInt16 fieldNum = TblGetRowID(table, row);

    *textH = SchemaFields[fieldNum].name;
    *textOffset = 0;
    *textAllocSize = MemHandleSize(SchemaFields[fieldNum].name);
    if (fld) {
        FieldAttrType attr;

        FldGetAttributes(fld, &attr);
        attr.singleLine = 1;
        attr.dynamicSize = 0;
        attr.underlined = 1;
        FldSetAttributes(fld, &attr);
        
        FldSetMaxChars(fld, 31);
    }

    return 0;
}           

static Boolean
SaveData(void * table, Int16 row, Int16 col)
{
    return false;
}

static UInt16 GetNumberOfRows(TablePtr table, UInt16 height)
{
	RectangleType rect;

	TblGetBounds(table, &rect);
	return rect.extent.y / height;
}

static void
LoadTable(TablePtr table)
{
    UInt16 numRows, row, fieldNum, i, numVisiblesRows;
    ListPtr list;
    FormPtr form;
    UInt16 height = FntLineHeight();
    form = FrmGetActiveForm();
    list = GetObjectPtr(form, ctlID_DesignView_PopupList);

#ifdef CONFIG_HANDERA
    if (FeatureSet.Handera) {
        VgaTableUseBaseFont(table, !VgaIsVgaFont(DBPrefs.TheStdFont));
    }
#endif

    numRows = TblGetNumberOfRows(table);
	numVisiblesRows = GetNumberOfRows(table, height);
    fieldNum = FirstField;
    for (row = 0; row < numRows; row++) {

        /* Stop the loop if no more fields are available. */
        if (fieldNum >= NumFields)
            break;
		if (row == numVisiblesRows)
			break;

        /* Mark the row as usable. */
        TblSetRowUsable(table, row, true);
        TblSetRowHeight(table, row, height);
        TblSetRowID(table, row, fieldNum);
        TblSetRowData(table, row, SchemaFields[fieldNum].uniqueID);
        TblSetRowSelectable(table, row, true);
        TblMarkRowInvalid(table, row);

        /* popup indicator */
        TblSetItemStyle(table, row, col_popup, customTableItem);
            
        /* old name */
        TblSetItemStyle(table, row, col_oldname, labelTableItem);
        TblSetItemPtr(table, row, col_oldname,
                          SchemaFields[fieldNum].old_name);
        if (FeatureSet.OS30)
            TblSetItemFont(table, row, col_oldname, stdFont);
        
        /* current name */
        TblSetItemStyle(table, row, col_name, textTableItem);
        if (FeatureSet.OS30)
            TblSetItemFont(table, row, col_name, stdFont);

        /* field type */
        TblSetItemStyle(table, row, col_type, popupTriggerTableItem);
        TblSetItemPtr(table, row, col_type, list);
        if (FeatureSet.OS30)
            TblSetItemFont(table, row, col_type, stdFont);
        i = 0;
        while (types[i].type != 0) {
            if (types[i].type == SchemaFields[fieldNum].type) {
                TblSetItemInt(table, row, col_type, types[i].index);
                break;
            }
            i++;
        }

        /* Advance the field number. */
        fieldNum++;
    }

    /* Mark the remainder of the rows as unusable. */
    while (row < numRows) {
        TblSetRowUsable(table, row, false);
        row++;
    }

    /* Update the scroll arrows. */
    UpdateScrollers();
}

static void
SetColumnWidths(TableType* table)
{
    Coord oldname_width, fieldtype_width;
    Int16 i, w;
    RectangleType bounds;
    FormPtr form;
    ListPtr lst;
    UInt16 lstIndex;

    oldname_width = FntCharsWidth(FieldString, StrLen(FieldString))+ FntCharsWidth(" XX:", 1);

    /* Determine and set the width of the field type column. */
    fieldtype_width = 0;
    for (i = 0; i < popup_strings_count; ++i) {
        w = FntCharsWidth(popup_strings[i], StrLen(popup_strings[i]));
        if (w > fieldtype_width)
            fieldtype_width = w;
    }

    /* The width of the field entry column is the remaining width. */
    TblGetBounds(table, &bounds);
    TblSetColumnWidth(table, col_popup, 10);
    /*on some OS a line appear after the col_popup when the line is deselected without this line*/
    TblSetColumnSpacing(table, col_popup, 0);
    TblSetColumnWidth(table, col_oldname, oldname_width);
    TblSetColumnWidth(table, col_type, fieldtype_width + 15);
    TblSetColumnWidth(table, col_name, bounds.extent.x - 10 - oldname_width
                      - (fieldtype_width + 15));

    form = FrmGetActiveForm();
    lstIndex = FrmGetObjectIndex(form, ctlID_DesignView_PopupList);
    lst = FrmGetObjectPtr(form, lstIndex);

    LstSetListChoices(lst, popup_strings, popup_strings_count);
    LstSetHeight(lst, popup_strings_count);

    FrmGetObjectBounds(form, lstIndex, &bounds);
    bounds.extent.x = fieldtype_width + 10;
    FrmSetObjectBounds(form, lstIndex, &bounds);
}

static void
InitTable(TablePtr table)
{
    UInt16 row, numRows;

    numRows = TblGetNumberOfRows(table);
    for (row = 0; row < numRows; row++) {
        TblSetRowUsable(table, row, false);
    }

    TblSetColumnUsable(table, col_popup, true);
    TblSetColumnUsable(table, col_oldname, true);
    TblSetColumnUsable(table, col_name, true);
    TblSetColumnUsable(table, col_type, true);

    TblSetCustomDrawProcedure(table, col_popup, DrawPopupIndicator);
    TblSetLoadDataProcedure(table, col_name, LoadData);
    TblSetSaveDataProcedure(table, col_name, SaveData);

    SetColumnWidths(table);

    LoadTable(table);
}

static void
Scroll(WinDirectionType direction)
{
    UInt16 newFirstField;

    newFirstField = FirstField;

    if (direction == winDown) {
        newFirstField++;
    } else {
        newFirstField--;
    }

    if (FirstField != newFirstField) {
        FormPtr form;
        TablePtr table;

        FirstField = newFirstField;

        form = FrmGetActiveForm();
        table = GetObjectPtr(form, ctlID_DesignView_Table);
        LoadTable(table);
        TblRedrawTable(table);
    }
}

static UInt32 *
CopyLocalSchema(DataSourceSchema * dst, SchemaDesignField * src, UInt16 numFields)
{
    UInt32 *reorder;
    UInt16 i;

    /* Flag an error if no fields were defined. */
    if (numFields == 0) {
        FrmAlert(alertID_NoFields);
        return NULL;
    }

    /* Build a schema from the current design to pass to the data source. */
    dst->numFields = numFields;
    dst->fields = AllocateMemPtr(numFields * sizeof(DataSourceFieldInfo));
    reorder = AllocateMemPtr(numFields * sizeof(UInt32));
    for (i = 0; i < numFields; ++i) {
        dst->fields[i].name = LockMemHandle(src[i].name);
        dst->fields[i].type = src[i].type;
        dst->fields[i].lock = src[i].lock;
        MemMove( &(dst->fields[i].data), &src[i].data, sizeof(DataSourceFieldData));
        reorder[i] = src[i].uniqueID;
    }

    return reorder;
}


static Boolean
UpdateNewDesign(void)
{
    DataSourceSchema schema;
    UInt32 *reorder;
    int i, numFields;
    void * key_data = NULL;
    UInt16 key_size;
    Err err;

    /* Determine the number of fields by finding the last nonempty field. */
    if (! DesignViewInitialSchema) {
        /* Blank schema, so search for number of fields to use. */
        for (i = NumFields - 1; i >= 0; i--) {
            char * p = LockMemHandle(SchemaFields[i].name);
            Int16 len = StrLen(p);
            UnlockMemPtr(p);

            if (len > 0) break;
        }
        numFields = i + 1;
    } else {
        /* Had an initial schema so just use the current number of fields. */
        numFields = NumFields;
    }

    //CopyLocalSchema return a pointer for reorder, we don't use it here
    if( (reorder = CopyLocalSchema( &schema, SchemaFields, numFields)))
        DeallocateMemPtr(reorder);
    else
        return false;

    /* Create the database. Creation routine will display any errors. */
    err = DesignViewDriver->class_ops.Create(DesignViewName, &schema,
                                             &key_data, &key_size);

    if (err == 0) {
        Chooser_RescanDatabases();
#if 0
        Chooser_AddDatabase(DesignViewName, DesignViewDriver->sourceID,
                            key_data, key_size);
#endif
    }
    if (key_data)
        DeallocateMemPtr(key_data);

    /* Unlock and free any temporary data. */
    /* with CopyLocalSchema there is not more memory allocated than the fields*/
    for (i = 0; i < numFields; ++i) {
        UnlockMemHandle(SchemaFields[i].name);
    }

    DeallocateMemPtr( schema.fields);

    if (err == 0) 
        return true;
    else
        return false;

    return true;
}

static Boolean
UpdateExistingDesign(void)
{
    Boolean changed, rebuild, result;
    DataSourceSchema schema;
    UInt32 *reorder;
    int i;

    reorder = CopyLocalSchema( &schema, SchemaFields, NumFields);
    if (!reorder)
        return false;
    
    /* Ask the data source about the extent of the changes. */
    CurrentSource->ops.CheckUpdateSchema(CurrentSource, &schema, reorder,
                                         &changed, &rebuild);

	
	if ( ForceRebuild ) {
		changed = rebuild = true;
	}

    result = true;
    if (changed) {
        if (rebuild) {
            /* The changes require a rebuild, inform the user. */
            if (FrmAlert(alertID_RebuildRequired) == 0) {
                void *context;

                context = ShowWorkingIndicator();
                CurrentSource = RebuildDatabase(CurrentSource, &schema, reorder);
                HideWorkingIndicator(context);
            } else {
                /* User does not want to do a rebuild. Stay in design mode. */
                result = false;
            }
        } else {
            /* User made changes that the data source can handle itself. */
            CurrentSource->ops.UpdateSchema(CurrentSource, &schema, reorder);
        }
    }

    // Unlock and free any temporary data. 
    // with CopyLocalSchema there is not more memory allocated than the fields
    for (i = 0; i < NumFields; ++i) {
        UnlockMemHandle(SchemaFields[i].name);
    }

    DeallocateMemPtr( schema.fields);
    DeallocateMemPtr(reorder);

    return result;
}

static UInt16 
CheckScripts() 
{
	UInt16 i;
	ScriptContext sc;
	CalculationData calc;

	// we do all the decompilation with this temporary data structure so
	// that the script's text and bytecode doesn't get messed up if the
	// compilation fails.
	MemSet( &calc, sizeof(CalculationData), 0);

	sc.src = NULL;
	sc.context = RECORD_FIELD_WRITING_CONTEXT;
	sc.flags = ScriptContext_DataAccess_DTC;
	sc.data_access.da.sdf = SchemaFields;
	sc.data_access.da.n = NumFields;


	for( i = 0; i < NumFields; i++ ) {
		sc.rec_field = i;

		if ( SchemaFields[i].type == FIELD_TYPE_CALCULATED ) {
			
			calc.scriptText = 
				DesignViewDriver->class_ops.DecompileScript(&SchemaFields[i].data.calc);			  

			if ( !DesignViewDriver->class_ops.CompileScript(&sc, &calc, true) ) {
				DeallocateMemHandle( calc.bytecode );
				DeallocateMemHandle( calc.scriptText );
				
				return i;
			}
			else {
				// the script may compile ok, but result in a different
				// bytecode than the last time it was compiled (e.g. if
				// we're now doing floating point arithmetic as opposed to
				// integer).
				if ( SchemaFields[i].data.calc.scriptText != 0 )
					DeallocateMemHandle(SchemaFields[i].data.calc.scriptText);
				if ( SchemaFields[i].data.calc.bytecode != 0 )
					DeallocateMemHandle(SchemaFields[i].data.calc.bytecode);

				MemMove( &SchemaFields[i].data.calc, &calc, sizeof( calc ) );

				// we don't want to actually copy the script text handle into
				// the SchemaFields, so erase it.
				SchemaFields[i].data.calc.scriptText = 0;
				DeallocateMemHandle( calc.scriptText );

				// we want to ensure that the bytecode is not deallocated
				// next time we call CompileScript with &calc, so erase that
				// reference to the handle.
				calc.bytecode = 0;



			}
		}
		
	} // for each field
	
	return SCRIPTS_ARE_COMPATIBLE;
}

static void
InsertField(UInt16 WhichField)
{
    Char * str1, *p;
    UInt16 i;

    /* Allocate more space for the additional field. */
    UnlockMemHandle(SchemaFieldsHandle);
    MemHandleResize(SchemaFieldsHandle,
                    (NumFields + 1) * sizeof(SchemaDesignField));
    SchemaFields = LockMemHandle(SchemaFieldsHandle);

    /* Move the fields after the insert position to the right. */
    if (WhichField < NumFields) {
        for (i = NumFields; i != WhichField; --i) {
            MemMove(&SchemaFields[i], &SchemaFields[i - 1],
                    sizeof(SchemaFields[i]));
        }
    }

    str1 = GetStringResource(stringID_New);

    SchemaFields[WhichField].name = AllocateMemHandle(StrLen(str1) + 1);
    p = LockMemHandle(SchemaFields[WhichField].name);
    StrCopy(p, str1);
    UnlockMemPtr(p);
    SchemaFields[WhichField].uniqueID = NextUniqueID++;
    SchemaFields[WhichField].type = FIELD_TYPE_STRING;
    SchemaFields[WhichField].lock = true;
    MemSet( &SchemaFields[WhichField].data, sizeof(DataSourceFieldData), 0);
    StrPrintF(SchemaFields[WhichField].old_name, FieldString, (Int16) WhichField);

    PutStringResource(str1);

    /* Increment the number of fields in the design. */
    NumFields++;
}

static void
DeleteField(UInt16 WhichField)
{
    Int16 i;

//    DeallocateMemHandle(SchemaFields[WhichField].name);
	ClearDesignSchemaField( WhichField );
    /* Move the other fields down. */
    for (i = WhichField; i < NumFields - 1; i++) {
        MemMove(&(SchemaFields[i]), &(SchemaFields[i + 1]),
                sizeof(SchemaFields[i]));
    }
    NumFields--;

    /* Reallocate the space in the handle to remove last position. */
    UnlockMemHandle(SchemaFieldsHandle);
    MemHandleResize(SchemaFieldsHandle,
                    NumFields * sizeof(SchemaDesignField));
    SchemaFields = LockMemHandle(SchemaFieldsHandle);
}

static void
HandleContextMenu(FormType* form, TableType* table, Int16 row)
{
    ListType* list;
    UInt16 list_index;
    UInt16 selection;
    RectangleType cell_bounds, list_bounds;
 
    /* Remove any selection UI that the table control does. */
    TblUnhighlightSelection(table);
 
    /* Retrieve the popup choices list. */
    list_index = FrmGetObjectIndex(form, ctlID_DesignView_InserterList);
    list = FrmGetObjectPtr(form, list_index);
 
    /* Set the location for the popup list. */
    TblGetItemBounds(table, row, col_popup, &cell_bounds);
    FrmGetObjectBounds(form, list_index, &list_bounds);
    list_bounds.topLeft.x = cell_bounds.topLeft.x;
    list_bounds.topLeft.y = cell_bounds.topLeft.y;
    FrmSetObjectBounds(form, list_index, &list_bounds);
 
    /* Set the choices which will be presented to the user. */
    LstSetSelection(list, noListSelection);
    /* Let the user choose an action. */
    selection = LstPopupList(list);

    /* Insert or delete a column as necessary. */
    if (selection != noListSelection) {
        UInt16 entry = TblGetRowID(table, row);

        switch (selection) {
        case 0:
            InsertField(entry);
            /* Redraw the table now that we have changed the ordering. */
            LoadTable(table);
            TblRedrawTable(table);
            break;
 
        case 1:
            InsertField(entry + 1);
            /* Redraw the table now that we have changed the ordering. */
            LoadTable(table);
            TblRedrawTable(table);
            break;
 
        case 2:
            DeleteField(entry);
            /* Redraw the table now that we have changed the ordering. */
            LoadTable(table);
            TblRedrawTable(table);
            break;

        case 3:
            ShowFieldProperties(entry);
            break;
        }
 
    }
} 

static void
FillLocalSchema(void)
{
    DataSourceSchema * schema;
    Int16 i, j;
    char * p;

    /* Determine the number of fields that we will display. */
    if (CurrentSource) {
        schema = &CurrentSource->schema;
        NumFields = schema->numFields;
    } else if (DesignViewInitialSchema) {
        schema = DesignViewInitialSchema;
        NumFields = schema->numFields;
    } else {
        schema = 0;
        NumFields = CONFIG_MAX_NUM_FIELDS_FOR_NEW_DB;
    }

    /* Allocate the temporary memory space. */
    SchemaFieldsHandle = AllocateMemHandle(NumFields * sizeof(SchemaDesignField));
    SchemaFields = LockMemHandle(SchemaFieldsHandle);
    NextUniqueID = 0;

    if (schema) {
        for (i = 0; i < NumFields; i++) {
            UInt32 len = StrLen(schema->fields[i].name);
            SchemaFields[i].name = AllocateMemHandle(1 + len);
            p = LockMemHandle(SchemaFields[i].name);
            StrCopy(p, schema->fields[i].name);
            UnlockMemPtr(p);
            SchemaFields[i].uniqueID = NextUniqueID++;
            SchemaFields[i].type = schema->fields[i].type;
            SchemaFields[i].lock = schema->fields[i].lock;
            SchemaFields[i].resetData = 0;
            switch (schema->fields[i].type){
            case FIELD_TYPE_STRING:
                if (schema->fields[i].data.string.defaultValue) {
                    SchemaFields[i].data.string.defaultValue = AllocateMemPtr( 33);
                    StrNCopy( SchemaFields[i].data.string.defaultValue, 
                        schema->fields[i].data.string.defaultValue, 32);
                } else {
                    SchemaFields[i].data.string.defaultValue = NULL;
                }
            break;
            case FIELD_TYPE_INTEGER:
                SchemaFields[i].data.integer.defaultValue = schema->fields[i].data.integer.defaultValue;
                SchemaFields[i].data.integer.increment = schema->fields[i].data.integer.increment;
            break;
#ifdef CONFIG_FLOAT
            case FIELD_TYPE_FLOAT:
                SchemaFields[i].data.floatNum.defaultValue = schema->fields[i].data.floatNum.defaultValue;
            break;
#endif
            case FIELD_TYPE_LIST:
            {
#ifdef TEST
                UInt8 *p, *q;
                UInt16 itemNum = schema->fields[i].data.list.numItems;
                SchemaFields[i].data.list.numItems = itemNum;
                SchemaFields[i].data.list.itemsH = AllocateMemHandle( MemHandleSize(schema->fields[i].data.list.itemsH));
                p = MemHandlLock(schema->fields[i].data.list.itemsH);
                q = MemHandlLock(SchemaFields[i].data.list.itemsH);
                MemMove(q, p, MemHandleSize(schema->fields[i].data.list.itemsH));
                UnlockMemPtr( p);
                UnlockMemPtr( q);
                SchemaFields[i].data.list.uniqID = AllocateMemPtr(MAX_LISTITEMS *sizeof(UInt16));
                MemSet( SchemaFields[i].data.list.uniqID, MAX_LISTITEMS * sizeof(UInt16), 0xFF);
                for (j = 0; j < itemNum; j++){
                    SchemaFields[i].data.list.uniqID[j] = j;
                }
#else
                UInt16 itemNum = schema->fields[i].data.list.numItems;
                SchemaFields[i].data.list.numItems = itemNum;
                SchemaFields[i].data.list.items = AllocateMemPtr( MAX_LISTITEMS * sizeof(char *));
                MemSet( SchemaFields[i].data.list.items, MAX_LISTITEMS * sizeof(char *), 0);
                SchemaFields[i].data.list.uniqID = AllocateMemPtr(MAX_LISTITEMS *sizeof(UInt16));
                MemSet( SchemaFields[i].data.list.uniqID, MAX_LISTITEMS * sizeof(UInt16), 0xFF);
                for (j = 0; j < itemNum; j++){
                    SchemaFields[i].data.list.items[j] = AllocateMemPtr( StrLen(schema->fields[i].data.list.items[j]) + 1);
                    StrCopy( SchemaFields[i].data.list.items[j], schema->fields[i].data.list.items[j]);
                    SchemaFields[i].data.list.uniqID[j] = j;
                }
#endif
            }
            break;
            case FIELD_TYPE_LINK:
                SchemaFields[i].data.link.dbID = schema->fields[i].data.link.dbID;
                SchemaFields[i].data.link.fieldNum = schema->fields[i].data.link.fieldNum;
            break;

            case FIELD_TYPE_LINKED:
                SchemaFields[i].data.linked.linkNum = schema->fields[i].data.linked.linkNum;
                SchemaFields[i].data.linked.fieldNum = schema->fields[i].data.linked.fieldNum;
            break;

#ifdef CONFIG_SCRIPTS
			case FIELD_TYPE_CALCULATED:
		    {
				UInt8 *src, *dst;

                MemMove( &(SchemaFields[i].data.calc), 
					   &(schema->fields[i].data.calc),
					   sizeof( CalculationData ) );

                if (schema->fields[i].data.calc.bytecode_sz) {
                    SchemaFields[i].data.calc.scriptText = 
    					DesignViewDriver->class_ops.DecompileScript(&schema->fields[i].data.calc);			  
    				SchemaFields[i].data.calc.bytecode = 
    					AllocateMemHandle( schema->fields[i].data.calc.bytecode_sz );
    				src = LockMemHandle( schema->fields[i].data.calc.bytecode );
    				dst = LockMemHandle( SchemaFields[i].data.calc.bytecode );
    				MemMove( dst, src, schema->fields[i].data.calc.bytecode_sz );
    				UnlockMemPtr( dst );
    				UnlockMemPtr( src );
                } else {
                    SchemaFields[i].data.calc.scriptText = AllocateMemHandle( 1);
                    SchemaFields[i].data.calc.bytecode = AllocateMemHandle( 1);
                }
			}
			break;
#endif

            default:
                MemMove( &(SchemaFields[i].data), &(schema->fields[i].data), sizeof(DataSourceFieldData));
            }
            SchemaFields[i].data.changed = false;
            StrPrintF(SchemaFields[i].old_name, FieldString, i);
        }
    } else {
        for (i = 0; i < NumFields; i++) {
            SchemaFields[i].name = AllocateMemHandle(33);
            p = LockMemHandle(SchemaFields[i].name);
            MemSet(p, 32, 0);
            UnlockMemPtr(p);
            SchemaFields[i].uniqueID = NextUniqueID++;
            SchemaFields[i].type = FIELD_TYPE_STRING;
            StrPrintF(SchemaFields[i].old_name, FieldString, i);
            MemSet( &SchemaFields[i].data, sizeof(DataSourceFieldData), 0);
        }
    }
}
static void
ClearDesignSchemaField( Int16 i)
{
    Int16 j;

	if ( i < 0 || i >= NumFields ) return;

	switch (SchemaFields[i].type){
	case FIELD_TYPE_STRING:
		if (SchemaFields[i].data.string.defaultValue)
			DeallocateMemPtr( SchemaFields[i].data.string.defaultValue);
        break;
	case FIELD_TYPE_DATE:
	case FIELD_TYPE_TIME:
	case FIELD_TYPE_INTEGER:
        break;            
	case FIELD_TYPE_LIST:
	{
#ifdef TEST
		if (SchemaFields[i].data.list.itemsH) {
			DeallocateMemHandle( SchemaFields[i].data.list.itemsH);
		}
		if (SchemaFields[i].data.list.uniqID)
			DeallocateMemPtr(SchemaFields[i].data.list.uniqID);
#else
		if (SchemaFields[i].data.list.items) {
			for (j = 0; j < SchemaFields[i].data.list.numItems; j++){
				if (SchemaFields[i].data.list.items[j])
					DeallocateMemPtr(SchemaFields[i].data.list.items[j]);
			}
			
			DeallocateMemPtr(SchemaFields[i].data.list.items);
		}
		if (SchemaFields[i].data.list.uniqID)
			DeallocateMemPtr(SchemaFields[i].data.list.uniqID);
#endif
	}
	break;
#ifdef CONFIG_SCRIPTS
	case FIELD_TYPE_CALCULATED:
	{
		if (SchemaFields[i].data.calc.scriptText) {
			DeallocateMemHandle(SchemaFields[i].data.calc.scriptText);
			SchemaFields[i].data.calc.scriptText = NULL;
		}
		if ( SchemaFields[i].data.calc.bytecode ) {
			DeallocateMemHandle(SchemaFields[i].data.calc.bytecode);
			SchemaFields[i].data.calc.bytecode = NULL;
		}				
	}
	break;
#endif
	default:
	}
	DeallocateMemHandle(SchemaFields[i].name);

}

static void
SetupPopupList(void)
{
    UInt16 i, index;

    /* Count the number of valid field types. */
    popup_strings_count = 0;
    i = 0;
    while (types[i].type != 0) {
        if (DesignViewDriver->class_ops.SupportsFieldType(types[i].type))
            popup_strings_count++;
        i++;
    }

    /* Allocate space for the popup list strings. */
    popup_strings = AllocateMemPtr(popup_strings_count * sizeof(char *));

    /* Fill in the popup list strings. */
    i = index = 0;
    while (types[i].type != 0) {
        if (DesignViewDriver->class_ops.SupportsFieldType(types[i].type)) {
            /* Copy the field type name string into place. */
            popup_strings[index] = CopyStringResource(types[i].nameID);
    
            /* Advance the index inside the popup_strings array. */
            types[i].index = index;
            index++;
        } else {
            types[i].index = -1;
        }
        i++;
    }
}

static Boolean
DesignViewUpdateForm(FormPtr form, UInt16 type)
{
    TablePtr table;

    if (type & updateResizeForm) {
        UInt16 dX, dY;

        UtilMoveResizeForm(form, &dX, &dY, true);

        if (dX || dY)
        {
            UtilMoveObject(form, ctlID_DesignView_DoneButton, 0, dY, true);
            UtilMoveObject(form, ctlID_DesignView_CancelButton, 0, dY, true);
            UtilMoveObject(form, ctlID_DesignView_UpButton, dX, dY, true);
            UtilMoveObject(form, ctlID_DesignView_DownButton, dX, dY, true);
            UtilResizeObject(form, ctlID_DesignView_Table, dX, dY, true);
        }
    }

    table = GetObjectPtr(form, ctlID_DesignView_Table);
    if (type & updateInitForm) {
        /* Initialize the table that holds the design information. */
        InitTable(table);
    }
	else
		LoadTable(table);

    /* Draw the form with field #1 appearing at the top. */
    FrmDrawForm(form);
    return true;
}

Boolean
DesignViewHandleEvent(EventPtr event)
{
    FormPtr form;
    FieldPtr fld;
    TablePtr table;
    Char * name;
    UInt16 i;

    switch (event->eType) {
    case frmOpenEvent:
        form = FrmGetActiveForm();

		FieldString = GetStringResource(stringID_FieldN);

        /* Initialize the working copy of the schema. */
        FillLocalSchema();
    
        /* Setup the popup list used for data types. */
        SetupPopupList();
    
        /* Set the title of depending on what mode we are in. */
        if (CurrentSource) {
            name = GetStringResource(stringID_DatabaseDesign);
        }
        else
            name = GetStringResource(stringID_NewDatabase);
        FrmCopyTitle(form, name);
        PutStringResource(name);
    
        /* Initialize the table that holds the design information. */
        FirstField = 0;
        if (FeatureSet.HiResP)
            PINSetInputAreaState(pinInputAreaOpen);
        DesignViewUpdateForm(form, updateInitForm | updateResizeForm);
		ForceRebuild = false;

        return true;

    case winDisplayChangedEvent:
        form = FrmGetActiveForm();
        DesignViewUpdateForm(form, updateResizeForm);
        return true;

    case frmUpdateEvent:
        form = FrmGetActiveForm();
        DesignViewUpdateForm(form, event->data.frmUpdate.updateCode);
        return true;

    case frmCloseEvent:
        form = FrmGetActiveForm();
        /* Make sure the table no longer has focus. */
        table = GetObjectPtr(form, ctlID_DesignView_Table);
        if (TblGetCurrentField(table)) {
            TblReleaseFocus(table);
        }
    
        /* Free the local working copy of the schema. */
        if (SchemaFields) {
            for( i = 0; i < NumFields; i++ )
				ClearDesignSchemaField(i);
			
			UnlockMemHandle(SchemaFieldsHandle);
			DeallocateMemHandle(SchemaFieldsHandle);
			SchemaFields = 0;
			SchemaFieldsHandle = 0;
			
        }
    
        /* Free the popup list strings. */
        if (popup_strings) {
            for( i = 0; i <popup_strings_count; i++)
                if( popup_strings[i])
                    DeallocateMemPtr(popup_strings[i]);
            DeallocateMemPtr(popup_strings);
            popup_strings = 0;
            popup_strings_count = 0;
        }

        /* Free the name selected by the creation dialog. */
        if (DesignViewName) {
            DeallocateMemPtr(DesignViewName);
            DesignViewName = 0;
        }

        /* Free any initial schema. */
        if (DesignViewInitialSchema) {
            FreeSchema(DesignViewInitialSchema);
            DeallocateMemPtr(DesignViewInitialSchema);
            DesignViewInitialSchema = 0;
        }

        PutStringResource( FieldString);

        break;

    case tblSelectEvent:
            if (event->data.tblSelect.column == col_popup) {
                HandleContextMenu(FrmGetActiveForm(),
                              event->data.tblSelect.pTable,
                              event->data.tblSelect.row);
            return true;
            } else if (event->data.tblSelect.column == col_type) {
                UInt16 fieldNum, index;

                fieldNum = TblGetRowID(event->data.tblSelect.pTable,
                                   event->data.tblSelect.row);
                index = TblGetItemInt(event->data.tblSelect.pTable,
                                                event->data.tblSelect.row,
                                                    event->data.tblSelect.column);
                i = 0;
				
                while (types[i].type != 0) {
					
                    if (types[i].index >= 0 && types[i].index == index) {
						Boolean deallocateFieldData = true;
						
						if ( SchemaFields[fieldNum].type == types[i].type ) {
							deallocateFieldData = false;
						}
						else {
							/** Free old data associated with field... */
							switch( SchemaFields[fieldNum].type) {
							case FIELD_TYPE_LIST:
								if( SchemaFields[fieldNum].data.list.items) {
									UInt8 j;
									
									for (j = 0; j < SchemaFields[fieldNum].data.list.numItems; j++) {
										DeallocateMemPtr(SchemaFields[fieldNum].data.list.items[j]);
									}
									DeallocateMemPtr( SchemaFields[fieldNum].data.list.items);
									SchemaFields[fieldNum].data.list.numItems = 0;
								}
								if (SchemaFields[fieldNum].data.list.uniqID)
									DeallocateMemPtr(SchemaFields[fieldNum].data.list.uniqID);
								break;
#ifdef CONFIG_SCRIPTS
							case FIELD_TYPE_CALCULATED:
								DeallocateMemHandle( SchemaFields[fieldNum].data.calc.scriptText );
								SchemaFields[fieldNum].data.calc.scriptText = NULL;
								break;                                       
#endif
							default:
							}

							MemSet( &SchemaFields[fieldNum].data, sizeof(DataSourceFieldData), 0);
							SchemaFields[fieldNum].type = types[i].type;

						}

						/** Allocate New Data Associated With Field */
                        switch( types[i].type) {
                        case FIELD_TYPE_LIST:
                            if( deallocateFieldData && !SchemaFields[fieldNum].data.list.items) {
                                SchemaFields[fieldNum].data.list.items = AllocateMemPtr( MAX_LISTITEMS*sizeof(char*));
                                SchemaFields[fieldNum].data.list.uniqID = AllocateMemPtr( MAX_LISTITEMS*sizeof(UInt16));
                                MemSet( SchemaFields[fieldNum].data.list.uniqID, MAX_LISTITEMS * sizeof(UInt16), 0xFF);
                                SchemaFields[fieldNum].data.list.numItems = 0;
                            }
                            ShowFieldProperties( fieldNum);
                        break;
#ifdef CONFIG_SCRIPTS
                        case FIELD_TYPE_CALCULATED:
							if ( deallocateFieldData ) 
								SchemaFields[fieldNum].data.calc.scriptText = AllocateMemHandle( const_AllForms_ScriptEditFieldLength );

                            ShowFieldProperties( fieldNum);
                        break;
#endif
                        case FIELD_TYPE_LINKED:
                            SchemaFields[fieldNum].data.linked.linkNum = 0xFFFF;
                            SchemaFields[fieldNum].data.linked.fieldNum = 0;
                        case FIELD_TYPE_LINK:
                             ShowFieldProperties( fieldNum);
                        break;

                        default:
                        }
                    }
                    i++;
                }
                return true;
            } else if (event->data.tblSelect.column == col_oldname) {
                /* A tap on the label edits the field name. */
                TblReleaseFocus(event->data.tblSelect.pTable);
                TblUnhighlightSelection(event->data.tblSelect.pTable);
                TblGrabFocus(event->data.tblSelect.pTable,
                                    event->data.tblSelect.row, col_name);
                fld = TblGetCurrentField(event->data.tblSelect.pTable);
                if (fld) {
                        FldGrabFocus(fld);
                        FldMakeFullyVisible(fld);
                }
                return true;
            }
        break;

    case ctlSelectEvent:
            switch (event->data.ctlSelect.controlID) {
            case ctlID_DesignView_DoneButton: {
				UInt16 bad_script;
				char num[4];

                if (CurrentSource) {
                        /* Discard all changes in read-only mode. */
                        if (CurrentSource->openMode == dmModeReadOnly) {
                            FrmGotoForm(formID_ListView);
                            return true;
                        }

						bad_script = CheckScripts();
						if ( bad_script != SCRIPTS_ARE_COMPATIBLE ) {
							StrIToA( num, bad_script );
							FrmCustomAlert(alertID_IncompatibleScriptError, 
										   num, " ", " " );
							return true;
						}

                        /* Try to implement any edits the user wants. */
                        if (! UpdateExistingDesign())
                            return true;
                } else {
						bad_script = CheckScripts();
						if ( bad_script != SCRIPTS_ARE_COMPATIBLE ) {
							StrIToA( num, bad_script );
							FrmCustomAlert(alertID_IncompatibleScriptError, 
										   num, " ", " " );
							return true;
						}

                        /* Try to create the database as specified. */
                        if (! UpdateNewDesign())
                            return true;
                }

			}
            case ctlID_DesignView_CancelButton:
                if (CurrentSource)
                        FrmGotoForm(formID_ListView);
                    else
                        FrmGotoForm(formID_Chooser);

                return true;
            }
        break;

    case ctlRepeatEvent:
            switch (event->data.ctlRepeat.controlID) {
            case ctlID_DesignView_UpButton:
                Scroll(winUp);
            break;

            case ctlID_DesignView_DownButton:
                Scroll(winDown);
            break;
            }
        break;

    case menuCmdBarOpenEvent:
        if (FeatureSet.OS35) {
            FieldPtr fld;
            
            form = FrmGetActiveForm();
            table = GetObjectPtr(form, ctlID_DesignView_Table);
            fld = TblGetCurrentField(table);
            if (fld)
                FldHandleEvent(fld, event);
            event->data.menuCmdBarOpen.preventFieldButtons = true;
        }
        return false;

    case menuEvent:
            switch (event->data.menu.itemID) {
            case menuitemID_Help:
                FrmHelp(stringID_Help_DesignView);
            return true;

            default:
            return HandleCommonMenuEvent(event->data.menu.itemID);
            }
        break;

    default:
        break;
    }

    return false;
}

static UInt16 FieldNumProperties;
static void
ShowFieldProperties(UInt16 whichField)
{
    FieldNumProperties = whichField;
    FrmPopupForm(formID_FieldProperties);
}

    static char **items;
    static UInt16 *itemsID;
    static int numItems;
    static char **items2;
    static int numItems2;
    static DataSource * source;
    static UInt16 lastItemSelected; // used by List & Calculated

static Boolean InitFieldPropertiesForm(FormPtr form, UInt16 *ObjectsTable) SECT_UI;
static Boolean Other_FieldPropertiesHandleEvent(EventPtr event) SECT_UI;
static Boolean String_FieldPropertiesHandleEvent(EventPtr event) SECT_UI;
static Boolean DateTime_FieldPropertiesHandleEvent(EventPtr event) SECT_UI;
static Boolean List_FieldPropertiesHandleEvent(EventPtr event) SECT_UI;
static Boolean Link_FieldPropertiesHandleEvent(EventPtr event) SECT_UI;
#ifdef CONFIG_SCRIPTS
static Boolean Calculated_FieldPropertiesHandleEvent(EventPtr event) SECT_UI;
#endif

Boolean
FieldPropertiesHandleEvent(EventPtr event)
{
    switch( SchemaFields[FieldNumProperties].type) {
    case FIELD_TYPE_STRING:
    case FIELD_TYPE_INTEGER:
    case FIELD_TYPE_FLOAT:
    return String_FieldPropertiesHandleEvent(event);

    case FIELD_TYPE_DATE:
    case FIELD_TYPE_TIME:
    return DateTime_FieldPropertiesHandleEvent(event);

    case FIELD_TYPE_LIST:
    return List_FieldPropertiesHandleEvent(event);


    case FIELD_TYPE_LINK:
    case FIELD_TYPE_LINKED:
    return Link_FieldPropertiesHandleEvent(event);

#ifdef CONFIG_SCRIPTS
    case FIELD_TYPE_CALCULATED:
    return Calculated_FieldPropertiesHandleEvent(event);
#endif

    default:
    return Other_FieldPropertiesHandleEvent(event);
    }
    return false;
}

static Boolean
InitFieldPropertiesForm(FormPtr form, UInt16 *ObjectsTable)
{
    ControlPtr ctl;
    UInt16 i = 0;

#ifdef CONFIG_HANDERA
        if (FeatureSet.Handera) {
            VgaFormModify(form, vgaFormModify160To240);
            if (!UtilCheckSizeForm(form, 0)) {
                FrmAlert(alertID_FormTooBig);
                FrmReturnToForm(0);
                return false;
            }
        }
#endif

        // Ste the box title
        FrmSetTitle(form, SchemaFields[FieldNumProperties].old_name);

        //set an unused control
        ctl = GetObjectPtr(form, ctlID_FieldProperties_LockBox);
        CtlSetValue(ctl, SchemaFields[FieldNumProperties].lock);

        //display all widget of the form
        while(ObjectsTable[i] != 0) {
#define ShowObject(f,id) FrmShowObject((f),FrmGetObjectIndex((f),(id)))
            ShowObject(form, ObjectsTable[i]);
#undef ShowObject

            //correct a bug from PalmOS 3.0 to 3.2
            if (!FeatureSet.OS35) {
                if (ObjectsTable[i] == ctlID_FieldProperties_DatabasesList)
                    ((ListPtr)FrmGetObjectPtr(form, FrmGetObjectIndex(form,
                                            ctlID_FieldProperties_DatabasesList)))->attr.usable = true;

                if (ObjectsTable[i] == ctlID_FieldProperties_FieldsList)
                    ((ListPtr)FrmGetObjectPtr(form, FrmGetObjectIndex(form,
                                            ctlID_FieldProperties_FieldsList)))->attr.usable = true;

                if (ObjectsTable[i] == ctlID_FieldProperties_LinkFieldList)
                    ((ListPtr)FrmGetObjectPtr(form, FrmGetObjectIndex(form,
                                            ctlID_FieldProperties_LinkFieldList)))->attr.usable = true;

                if (ObjectsTable[i] == ctlID_FieldProperties_ItemList)
                    ((ListPtr)FrmGetObjectPtr(form, FrmGetObjectIndex(form,
                                            ctlID_FieldProperties_ItemList)))->attr.usable = true;
            }

            i++;
        }
    return true;
}

static Boolean
Other_FieldPropertiesHandleEvent(EventPtr event)
{
    FormPtr form;

    switch (event->eType) {
    case frmOpenEvent:
        form = FrmGetActiveForm();
#ifdef CONFIG_HANDERA
        if (FeatureSet.Handera) {
            VgaFormModify(form, vgaFormModify160To240);
            if (!UtilCheckSizeForm(form, 0)) {
                FrmAlert(alertID_FormTooBig);
                FrmReturnToForm(0);
                return false;
            }
        }
#endif

        //draw the form
        FrmDrawForm(form);
    return true;
    case frmCloseEvent:
    return false;
    case ctlSelectEvent:
        switch (event->data.ctlSelect.controlID) {
        case ctlID_FieldProperties_OK:
            form = FrmGetActiveForm();
            {
                ControlPtr ctl;
                ctl = GetObjectPtr(form, ctlID_FieldProperties_LockBox);
                SchemaFields[FieldNumProperties].lock = CtlGetValue(ctl);
            }
        //continue to free the form

        case ctlID_FieldProperties_Cancel:
            FrmReturnToForm(0);
        return true;
        default:
        }
    break;
    default:
    }
    return false;
}

static Boolean
String_FieldPropertiesHandleEvent(EventPtr event)
{
    Boolean closeAll = false;
    FormPtr form;
    FieldPtr fld;
    MemHandle h;
    UInt16 StringObjectsTable[] =
    {
        9002,
        ctlID_FieldProperties_DefaultValueField,
        0
    };
    UInt16 IntegerObjectsTable[] =
    {
        9002,
		9003,
        ctlID_FieldProperties_DefaultValueField,
        ctlID_FieldProperties_IncrementField,
        0
    };


    switch (event->eType) {
    case frmOpenEvent: {
        Char *s;

        form = FrmGetActiveForm();

        //init the form
        fld = GetObjectPtr(form, ctlID_FieldProperties_DefaultValueField);
        h = AllocateMemHandle( 33);
        s = LockMemHandle(h);
        switch ( SchemaFields[FieldNumProperties].type) {
        case FIELD_TYPE_STRING: {
            if (!InitFieldPropertiesForm(form, StringObjectsTable))
                return true;

            if (SchemaFields[FieldNumProperties].data.string.defaultValue) {
                StrNCopy( s, SchemaFields[FieldNumProperties].data.string.defaultValue, 32);
				SchemaFields[FieldNumProperties].data.string.defaultValue[31] = 0;
			}
            else
                s[0] = 0;
            }
        break;
        case FIELD_TYPE_INTEGER: {
            Int32 value = SchemaFields[FieldNumProperties].data.integer.defaultValue;

            if (!InitFieldPropertiesForm(form, IntegerObjectsTable))
                return true;

            MemSet(s, 32, 0);
            StrIToA(s, value);
            }
        break;
#ifdef CONFIG_FLOAT
        case FIELD_TYPE_FLOAT: {
            double fvalue = SchemaFields[FieldNumProperties].data.floatNum.defaultValue;

            if (!InitFieldPropertiesForm(form, StringObjectsTable))
                return true;

            MemSet(s, 32, 0 );
            DoublePrintValue( fvalue, s, kMaxMantissaDigits);
            }
        break;
#endif
        default:
        }
        UnlockMemHandle(h);
        FldSetTextHandle( fld, h );

        if (SchemaFields[FieldNumProperties].type == FIELD_TYPE_INTEGER) {
            Int32 value = (Int32)SchemaFields[FieldNumProperties].data.integer.increment;
            fld = GetObjectPtr(form, ctlID_FieldProperties_IncrementField);
            h = AllocateMemHandle( 33);
            s = LockMemHandle(h);
            MemSet(s, 32, 0);
            StrIToA(s, value);
            UnlockMemHandle(h);
            FldSetTextHandle( fld, h );
        }

        //draw the form
        FrmDrawForm(form);
      }
    return true;
    case frmCloseEvent:
        closeAll = true;
    goto freeform_String_FieldProperties;
    case ctlSelectEvent:
        switch (event->data.ctlSelect.controlID) {
        case ctlID_FieldProperties_OK:
            form = FrmGetActiveForm();

            fld = GetObjectPtr(form, ctlID_FieldProperties_DefaultValueField);

            switch ( SchemaFields[FieldNumProperties].type) {
            case FIELD_TYPE_STRING:
                if (FldGetTextLength(fld)) {

                    if (SchemaFields[FieldNumProperties].data.string.defaultValue)
                        DeallocateMemPtr(SchemaFields[FieldNumProperties].data.string.defaultValue);

                    SchemaFields[FieldNumProperties].data.string.defaultValue = AllocateMemPtr( FldGetTextLength(fld) + 1);
                    StrNCopy( SchemaFields[FieldNumProperties].data.string.defaultValue, FldGetTextPtr(fld), 32);
                    if (SchemaFields[FieldNumProperties].data.string.defaultValue[0] != '\0')
                        SchemaFields[FieldNumProperties].data.link.changed = true;

                }
            break;
            case FIELD_TYPE_INTEGER:
                SchemaFields[FieldNumProperties].data.integer.defaultValue = String2Long(FldGetTextPtr(fld));

                fld = GetObjectPtr(form, ctlID_FieldProperties_IncrementField);
                SchemaFields[FieldNumProperties].data.integer.increment = (Int16)String2Long(FldGetTextPtr(fld));

                if (SchemaFields[FieldNumProperties].data.integer.defaultValue != 0 
                        || SchemaFields[FieldNumProperties].data.integer.increment != 0)
                    SchemaFields[FieldNumProperties].data.link.changed = true;
            break;
#ifdef CONFIG_FLOAT
            case FIELD_TYPE_FLOAT:
                if (FldGetTextPtr(fld) && FldGetTextPtr(fld)[0] != 0)
                    DoubleAToD( &SchemaFields[FieldNumProperties].data.floatNum.defaultValue, FldGetTextPtr(fld));
                else
                    SchemaFields[FieldNumProperties].data.integer.defaultValue = 0.0;
                if (SchemaFields[FieldNumProperties].data.integer.defaultValue != 0.0)
                    SchemaFields[FieldNumProperties].data.link.changed = true;
            break;
#endif
            default:
            }

            {
                ControlPtr ctl;
                ctl = GetObjectPtr(form, ctlID_FieldProperties_LockBox);
                SchemaFields[FieldNumProperties].lock = CtlGetValue(ctl);
            }
        //continue to free the form

        case ctlID_FieldProperties_Cancel:
        goto freeform_String_FieldProperties;
        default:
        }
    break;
    default:
    }
    return false;

freeform_String_FieldProperties:
        form = FrmGetActiveForm();
        fld = GetObjectPtr(form, ctlID_FieldProperties_DefaultValueField);
        h = FldGetTextHandle(fld);
        FldSetTextHandle( fld, NULL );
        DeallocateMemHandle( h);

        if (SchemaFields[FieldNumProperties].type == FIELD_TYPE_INTEGER) {
            fld = GetObjectPtr(form, ctlID_FieldProperties_IncrementField);
            h = FldGetTextHandle(fld);
            FldSetTextHandle( fld, NULL );
            DeallocateMemHandle( h);
        }
        if (closeAll)
            return false;
        else {
            FrmReturnToForm(0);
            return true;
        }
}


static Boolean
DateTime_FieldPropertiesHandleEvent(EventPtr event)
{
    FormPtr form;
    ControlPtr ctl;	
	Char *s;

    UInt16 ObjectsTable[] =
    {
        9002,
		ctlID_FieldProperties_DateTimeTrigger,
        ctlID_FieldProperties_DateTimeNow,
        0
    };

    switch (event->eType) {
    case frmOpenEvent:
      {

        form = FrmGetActiveForm();
        if (!InitFieldPropertiesForm(form, ObjectsTable))
            return true;

		s = AllocateMemPtr( 32 );

		// dtcommon is a pretend structure that contains common
		// information from date and time
		switch ( SchemaFields[FieldNumProperties].data.dtcommon.validDefault ) {
		case NOW_DEFAULT:
			ctl = GetObjectPtr( form, ctlID_FieldProperties_DateTimeNow );
			CtlSetValue( ctl, true );

			// fall through
			
		case INVALID_DEFAULT:
			StrCopy( s, "N/A" );
			break;
			
		case CONSTANT_DEFAULT:
			if ( SchemaFields[FieldNumProperties].type == FIELD_TYPE_TIME) {
				TimeToAscii( SchemaFields[FieldNumProperties].data.time.defaultValue.hour,
							 SchemaFields[FieldNumProperties].data.time.defaultValue.min,
							 PrefGetPreference(prefTimeFormat),
							 s);
			}
			else {
				DateToAscii( SchemaFields[FieldNumProperties].data.date.defaultValue.month,
							 SchemaFields[FieldNumProperties].data.date.defaultValue.day,
							 SchemaFields[FieldNumProperties].data.date.defaultValue.year,
							 PrefGetPreference(prefLongDateFormat),
							 s);				
			}
			break;
		}

		ctl = GetObjectPtr( form, ctlID_FieldProperties_DateTimeTrigger );
		CtlSetLabel( ctl, s );


        FrmDrawForm(form);
      }
	  return true;
    case frmCloseEvent:
        form = FrmGetActiveForm();
		ctl = GetObjectPtr( form, ctlID_FieldProperties_DateTimeTrigger );
		DeallocateMemPtr(  (Char *)CtlGetLabel( ctl ) );

		return false;

    case ctlSelectEvent:
		form = FrmGetActiveForm();

        switch (event->data.ctlSelect.controlID) {
        case ctlID_FieldProperties_DateTimeTrigger:
			if ( SchemaFields[FieldNumProperties].type == FIELD_TYPE_DATE ) {

				Boolean redraw = false;
				Int16 month,day, year;
				DateTimeType datetime;
				ListPtr lst;
				DateData *date = &SchemaFields[FieldNumProperties].data.date;

				lst = GetObjectPtr( form, ctlID_FieldProperties_DateTimeChoiceList );

				if ( LstPopupList( lst ) == 0 ) {

					if ( date->validDefault == INVALID_DEFAULT || 
						 date->validDefault == NOW_DEFAULT ) {
						
						TimSecondsToDateTime( TimGetSeconds(), &datetime );
						month = datetime.month;
						year = datetime.year;
						day = datetime.day;
					}
					else {

						month = date->defaultValue.month;
						day = date->defaultValue.day;
						year = date->defaultValue.year;
					}
					
					if ( SelectDay(selectDayByDay, &month, &day, &year, "Edit Date") ) {
						date->validDefault = CONSTANT_DEFAULT;
						date->defaultValue.month = month;
						date->defaultValue.day = day;
						date->defaultValue.year = year;						
						redraw = true;
					}



				}
				else {
					date->validDefault = INVALID_DEFAULT;
					redraw = true;
				}

				if ( redraw ) {
					
					s = AllocateMemPtr( 32 );
					if ( date->validDefault == CONSTANT_DEFAULT ) {
						DateToAscii( date->defaultValue.month,
									 date->defaultValue.day,
									 date->defaultValue.year,
									 PrefGetPreference(prefLongDateFormat),
									 s);
					}
					else 
						StrCopy( s, "N/A" );

					
					ctl = GetObjectPtr( form, ctlID_FieldProperties_DateTimeNow );
					CtlSetValue( ctl, false );
					CtlDrawControl( ctl );

					ctl = GetObjectPtr( form, ctlID_FieldProperties_DateTimeTrigger );
		
					DeallocateMemPtr( (Char *) CtlGetLabel( ctl ) );
					CtlSetLabel( ctl, s );
					CtlDrawControl( ctl );
				}
			}
			else if ( SchemaFields[FieldNumProperties].type == FIELD_TYPE_TIME ) {
				
				Boolean redraw = false;
				ListPtr lst;
				TimeData *time = &SchemaFields[FieldNumProperties].data.time;

				lst = GetObjectPtr( form, ctlID_FieldProperties_DateTimeChoiceList );


				if( LstPopupList( lst ) == 0 ) {
					if ( FeatureSet.OS31 ) {
						Int16 hour, minute;
						DateTimeType datetime;
						
						if ( time->validDefault == INVALID_DEFAULT || 
							 time->validDefault == NOW_DEFAULT ) {
							
							TimSecondsToDateTime( TimGetSeconds(), &datetime );
							hour = datetime.hour;
							minute = datetime.minute;
						}
						else {
							hour = time->defaultValue.hour;
							minute = time->defaultValue.min;
						}

						if ( SelectOneTime( &hour, &minute, "Edit Time" )) {
							time->validDefault = CONSTANT_DEFAULT;
							time->defaultValue.hour = hour;
							time->defaultValue.min = minute;
							redraw = true;

						}
					}
					else {
						TimeType start, finish;
						DateTimeType datetime;

						if ( time->validDefault == INVALID_DEFAULT ||
							 time->validDefault == NOW_DEFAULT ) {
							
							TimSecondsToDateTime( TimGetSeconds(), &datetime );
							start.hours = datetime.hour;
							start.minutes = datetime.minute;
						}
						else {
							start.hours = time->defaultValue.hour;
							start.minutes = time->defaultValue.min;
						}
						

						finish.hours = noTime;
						finish.minutes = noTime;
						
						if (SelectTimeV33(&start, &finish, false,
										  "Edit Time", 0)){
							
							time->validDefault = CONSTANT_DEFAULT;
							time->defaultValue.hour = start.hours;
							time->defaultValue.min = start.minutes;
							ctl = GetObjectPtr( form, ctlID_FieldProperties_DateTimeNow );
							CtlSetValue( ctl, false );
							CtlDrawControl( ctl );
							redraw = true;
						}
					}
					
				}
				else {
					time->validDefault = INVALID_DEFAULT;
					redraw = true;
				}
				if ( redraw ) {
					
					s = AllocateMemPtr( 32 );
					if ( time->validDefault ) {
						TimeToAscii( time->defaultValue.hour,
									 time->defaultValue.min,
									 PrefGetPreference(prefTimeFormat),
									 s);
					}
					else
						StrCopy( s, "N/A" );					
					
					ctl = GetObjectPtr( form, ctlID_FieldProperties_DateTimeNow );
					CtlSetValue( ctl, false );
					CtlDrawControl( ctl );
					
					ctl = GetObjectPtr( form, ctlID_FieldProperties_DateTimeTrigger );
					
					DeallocateMemPtr( (Char *) CtlGetLabel( ctl ) );

					CtlSetLabel( ctl, s );
					CtlDrawControl( ctl );
				}
			
			}
			break;
			
        case ctlID_FieldProperties_DateTimeNow:
            form = FrmGetActiveForm();

			ctl = GetObjectPtr( form, ctlID_FieldProperties_DateTimeNow );
			if ( CtlGetValue( ctl ) ) {
				SchemaFields[FieldNumProperties].data.dtcommon.validDefault = NOW_DEFAULT;

				ctl = GetObjectPtr( form, ctlID_FieldProperties_DateTimeTrigger );
				s = AllocateMemPtr( 32 );
				StrCopy( s, "N/A" );
				DeallocateMemPtr( (Char *) CtlGetLabel(ctl) );
				CtlSetLabel( ctl, s );
				CtlDrawControl( ctl );
				
			}
			// this is  valid for both data.date and data.time
			else SchemaFields[FieldNumProperties].data.dtcommon.validDefault = INVALID_DEFAULT;

		return true;

        case ctlID_FieldProperties_OK:
			SchemaFields[FieldNumProperties].data.dtcommon.changed = true;
			
			// fall through
        case ctlID_FieldProperties_Cancel:
            form = FrmGetActiveForm();
			ctl = GetObjectPtr( form, ctlID_FieldProperties_DateTimeTrigger );
			DeallocateMemPtr( (Char *) CtlGetLabel(ctl) );


            FrmReturnToForm(0);
		return true;
        default:
        }
    break;
    default:
    }
    return false;
}

static Boolean
List_FieldPropertiesHandleEvent(EventPtr event)
{
    FormPtr form;
    FieldPtr fld;
    ListPtr lst;
    UInt16 i;
    UInt16 ObjectsTable[] =
    {
        ctlID_FieldProperties_ItemList,
        ctlID_FieldProperties_NewItemButton,
        ctlID_FieldProperties_ItemField,
        ctlID_FieldProperties_DeleteItemButton,
        0
    };

    switch (event->eType) {
    case frmOpenEvent:
        items =  NULL;
        numItems = 0;
        itemsID = NULL;

        form = FrmGetActiveForm();
        if (!InitFieldPropertiesForm(form, ObjectsTable))
            return true;

        items = SchemaFields[FieldNumProperties].data.list.items;
        itemsID = SchemaFields[FieldNumProperties].data.list.uniqID;
        numItems = SchemaFields[FieldNumProperties].data.list.numItems;
        lastItemSelected = -1;

        FrmDrawForm(form);
        FrmUpdateForm(formID_FieldProperties, 0);
    return true;
    case frmCloseEvent:
        form = FrmGetActiveForm();
        lst = GetObjectPtr(form, ctlID_FieldProperties_ItemList);
        LstSetListChoices( lst, NULL, 0);
    return false;
    case frmUpdateEvent:
        form = FrmGetActiveForm();
        lst = GetObjectPtr(form, ctlID_FieldProperties_ItemList);
        LstSetListChoices( lst, items, numItems);
        LstSetTopItem( lst, event->data.frmUpdate.updateCode);
        LstSetSelection( lst, -1);
        LstDrawList(lst);
    return true;
    case lstSelectEvent:
        switch (event->data.lstSelect.listID) {
        case ctlID_FieldProperties_ItemList:
            form = FrmGetActiveForm();
            lst = GetObjectPtr(form, ctlID_FieldProperties_ItemList);
            if (lastItemSelected == event->data.lstSelect.selection) {
                LstSetSelection( lst, -1);
                lastItemSelected = -1;
                return true;
            }
            lastItemSelected = event->data.lstSelect.selection;
        break;
        default:
        }
    break;
    case ctlSelectEvent:
        switch (event->data.ctlSelect.controlID) {
        case ctlID_FieldProperties_NewItemButton:
            if (numItems > MAX_LISTITEMS-1)
			  return true;
            form = FrmGetActiveForm();
            fld = GetObjectPtr(form, ctlID_FieldProperties_ItemField);
            if (( i = FldGetTextLength(fld))) {
                UInt16 itemNum;
                UInt16 tempoNumItems = numItems;

                if (!items) {
                    items = AllocateMemPtr(MAX_LISTITEMS*sizeof(char*));
                    numItems = 0;
                }

                lst = GetObjectPtr(form, ctlID_FieldProperties_ItemList);
                itemNum = LstGetSelection(lst);
                if (itemNum == noListSelection)
                    itemNum = numItems;
                while( itemNum < tempoNumItems) {
                    items[tempoNumItems] = items[tempoNumItems-1];
                    itemsID[tempoNumItems] = itemsID[tempoNumItems-1];
                    tempoNumItems--;
                }

                items[itemNum] = AllocateMemPtr(FldGetTextLength(fld) + 1);
                itemsID[itemNum] = numItems;
                StrCopy( items[itemNum], FldGetTextPtr(fld));

                numItems++;
                SchemaFields[FieldNumProperties].data.list.changed = true;

                FrmUpdateForm(formID_FieldProperties, itemNum);
                // we need this for palm os 3.1, otherwise we crash
                FldSetSelection(fld, 0, StrLen( items[itemNum] ) );
            }
        return true;
        case ctlID_FieldProperties_DeleteItemButton:
        {
            UInt16 itemNum;
            if (numItems == 0)
                    return true;
            form = FrmGetActiveForm();
            lst = GetObjectPtr(form, ctlID_FieldProperties_ItemList);
            itemNum = LstGetSelection(lst);
            if (itemNum == noListSelection)
                return true;
            numItems--;
            DeallocateMemPtr(items[itemNum]);
            if (numItems) {
                while( itemNum < numItems) {
                    items[itemNum] = items[itemNum+1];
                    itemsID[itemNum] = itemsID[itemNum+1];
                    itemNum++;
                }
                items[numItems] = items[numItems+1];
                itemsID[numItems] = itemsID[numItems+1];
            } else {
                DeallocateMemPtr(items);
                items = NULL;
            }
            SchemaFields[FieldNumProperties].data.list.changed = true;

            FrmUpdateForm(formID_FieldProperties, itemNum);
        }
        return true;
        case ctlID_FieldProperties_OK:
            form = FrmGetActiveForm();
            SchemaFields[FieldNumProperties].data.list.numItems = numItems;
            SchemaFields[FieldNumProperties].data.list.items = items;
            SchemaFields[FieldNumProperties].data.list.uniqID = itemsID;
            {
                ControlPtr ctl;
                ctl = GetObjectPtr(form, ctlID_FieldProperties_LockBox);
                SchemaFields[FieldNumProperties].lock = CtlGetValue(ctl);
            }
        case ctlID_FieldProperties_Cancel:
            form = FrmGetActiveForm();
            lst = GetObjectPtr(form, ctlID_FieldProperties_ItemList);
            LstSetListChoices( lst, NULL, 0);
            FrmReturnToForm(0);
        return true;
        default:
        }
    break;
    default:
    }
    return false;
}

static Boolean
Link_FieldPropertiesHandleEvent(EventPtr event)
{
    Boolean closeAll;
    FormPtr form;
    ListPtr lst;
    ControlPtr ctl;
    ChooserRecord * record;
    MemHandle h;
    Char * p;
    UInt16 i, j;
    UInt16 objectsTable[] =
    {
        ctlID_FieldProperties_DatabasesList,
        ctlID_FieldProperties_FieldsList,
        ctlID_FieldProperties_KeepLinksBox,
        0
    };

    switch (event->eType) {
    case frmOpenEvent:
        items =  NULL;
        numItems = 0;
        itemsID = NULL;
        source = 0;
        items2 = NULL;
        numItems2 = 0;

        form = FrmGetActiveForm();

        Chooser_Open();

        switch ( SchemaFields[FieldNumProperties].type) {
        case FIELD_TYPE_LINK: {
            UInt16 databaseRecord = dmMaxRecordIndex;

            if (!InitFieldPropertiesForm(form, objectsTable))
                return true;

            while (Chooser_Seek( &databaseRecord, 1, dmSeekBackward)) numItems++;

            if (databaseRecord != dmMaxRecordIndex) numItems++;
            items = AllocateMemPtr( numItems * sizeof( char *));
            databaseRecord = 0;
            for (i = 0; i < numItems; i++) {

                if (!Chooser_Seek( &databaseRecord, 0, dmSeekForward))
                    break;
                h = DmQueryRecord(Chooser_Ref(), databaseRecord);
                record = LockMemHandle(h);
                items[i] = AllocateMemPtr( 33);
                StrCopy(items[i], record->title);
                UnlockMemHandle(h);
                databaseRecord++;
            }
            lst = GetObjectPtr(form, ctlID_FieldProperties_DatabasesList);
            LstSetListChoices( lst, items, numItems);
            LstSetSelection( lst, noListSelection);
            lst = GetObjectPtr(form, ctlID_FieldProperties_FieldsList);
            LstSetSelection( lst, noListSelection);

            }
        break;
        case FIELD_TYPE_LINKED:
            if (!InitFieldPropertiesForm(form, objectsTable))
                return true;

            for (i = 0; i < FieldNumProperties; i++)
                if (SchemaFields[i].type == FIELD_TYPE_LINK)
                    numItems++;

            if (!numItems) {
                FrmAlert(alertID_CreateLinkField);
            } else {
                items = AllocateMemPtr( numItems * sizeof( char *));
                for (i = 0, j = 0; i < FieldNumProperties; i++) {
                    if (SchemaFields[i].type == FIELD_TYPE_LINK) {
                        items[j] = AllocateMemPtr(33);
                        p = LockMemHandle(SchemaFields[i].name);
                        StrNCopy(items[j++], p, 32);
                        UnlockMemPtr(p);
                    }
                }
                lst = GetObjectPtr(form, ctlID_FieldProperties_DatabasesList);
                LstSetListChoices( lst, items, numItems);
                LstSetSelection( lst, noListSelection);
				lst = GetObjectPtr(form, ctlID_FieldProperties_FieldsList);
				LstSetSelection( lst, noListSelection);

            }
        break;
        default:
        }

        FrmDrawForm(form);
    return true;
    case frmCloseEvent:
        closeAll = true;
    goto freeform_Link_FieldProperties;
    case lstSelectEvent:
        switch (event->data.lstSelect.listID) {
        case ctlID_FieldProperties_DatabasesList: {
            UInt16 index;
            ChooserRecord * record;

            switch ( SchemaFields[FieldNumProperties].type) {
            case FIELD_TYPE_LINKED: {
                i = j = 0;

                while (i < FieldNumProperties)
                    if (SchemaFields[i++].type == FIELD_TYPE_LINK && event->data.lstSelect.selection == j++)
                        break;

                if (DmFindRecordByID(Chooser_Ref(), SchemaFields[i-1].data.link.dbID, &index)){
                    return true;
                }
                }
            break;
            case FIELD_TYPE_LINK:
                index = event->data.lstSelect.selection;
            break;
            default:
            }

            h = DmQueryRecord(Chooser_Ref(), index);

            record = LockMemHandle(h);

            if (source)
                CloseDataSource( source);

            source = OpenDataSource(DriverList, dmModeReadOnly,
                                    record->sourceID,
                                    (void *) (record + 1),
                                    MemHandleSize(h) -
                                    sizeof(ChooserRecord));
            UnlockMemHandle(h);

            if (!source)
                break;

/*Marc bug correction*/
            if (items2) {
                DeallocateMemPtr( items2);
                items2 = NULL;
            }

            numItems2 = source->schema.numFields;
            if (numItems2)
                items2 = AllocateMemPtr( numItems2 * sizeof( char *));

            for( i = 0; i < numItems2; i++) {
                    items2[i] = source->schema.fields[i].name;
            }
            form = FrmGetActiveForm();
            lst = GetObjectPtr(form, ctlID_FieldProperties_FieldsList);
            LstSetListChoices( lst, items2, numItems2);
            LstSetSelection( lst, noListSelection);
            LstDrawList(lst);
          }
        return true;
        default:
        }
    break;
    case ctlSelectEvent:
        switch (event->data.ctlSelect.controlID) {
        case ctlID_FieldProperties_LockBox:
            form = FrmGetActiveForm();
            ctl = GetObjectPtr(form, ctlID_FieldProperties_LockBox);

            SchemaFields[FieldNumProperties].lock = CtlGetValue(ctl);
        return true;
        case ctlID_FieldProperties_OK:
          {
            UInt16  listSelection;

            form = FrmGetActiveForm();
            lst = GetObjectPtr(form, ctlID_FieldProperties_DatabasesList);
            listSelection = LstGetSelection( lst);
            switch ( SchemaFields[FieldNumProperties].type) {
            case FIELD_TYPE_LINK:
                if ( listSelection != noListSelection) {
                    UInt32 uniqID;

                    DmRecordInfo(Chooser_Ref(), listSelection, NULL,
                                &uniqID, NULL);
                    SchemaFields[FieldNumProperties].data.link.dbID = uniqID;
                    lst = GetObjectPtr(form, ctlID_FieldProperties_FieldsList);
                    listSelection = LstGetSelection( lst);
                    if (listSelection == noListSelection) listSelection = 0;
                    SchemaFields[FieldNumProperties].data.link.fieldNum = listSelection;
                    SchemaFields[FieldNumProperties].data.link.changed = true;
                } else {
                    SchemaFields[FieldNumProperties].data.link.dbID = 0;
                    SchemaFields[FieldNumProperties].data.link.fieldNum = 0;
                    SchemaFields[FieldNumProperties].data.link.changed = true;
                }
            break;
            case FIELD_TYPE_LINKED:
                if ( listSelection != noListSelection) {
                    j = i = 0;
                    while (i < FieldNumProperties)
                       if (SchemaFields[i++].type == FIELD_TYPE_LINK && listSelection == j++)
                           break;
 
                    SchemaFields[FieldNumProperties].data.linked.linkNum = i - 1;
                    lst = GetObjectPtr(form, ctlID_FieldProperties_FieldsList);
                    listSelection = LstGetSelection( lst);
                    if (listSelection == noListSelection) listSelection = 0;

                    SchemaFields[FieldNumProperties].data.linked.fieldNum = listSelection;
                    SchemaFields[FieldNumProperties].data.linked.changed = true;
                } else {
                    SchemaFields[FieldNumProperties].data.linked.linkNum = 0xFFFF;
                    SchemaFields[FieldNumProperties].data.linked.fieldNum = 0;
                    SchemaFields[FieldNumProperties].data.linked.changed = true;
                }
            break;
            default:
            }
/*
         Not currently used, and it breaks if no source is opened.
            SchemaFields[FieldNumProperties].data.link.returnType = 
                source->schema.fields[SchemaFields[FieldNumProperties].data.link.fieldNum].type;
*/
            ctl = GetObjectPtr(form, ctlID_FieldProperties_KeepLinksBox);
            SchemaFields[FieldNumProperties].resetData = CtlGetValue(ctl);

            {
                ControlPtr ctl;
                ctl = GetObjectPtr(form, ctlID_FieldProperties_LockBox);
                SchemaFields[FieldNumProperties].lock = CtlGetValue(ctl);
            }
          }
        case ctlID_FieldProperties_Cancel:
            closeAll = false;
        goto freeform_Link_FieldProperties;
        default:
        }
    break;
    default:
    }
    return false;

freeform_Link_FieldProperties:
        form = FrmGetActiveForm();
        lst = GetObjectPtr(form, ctlID_FieldProperties_DatabasesList);
        LstSetListChoices( lst, NULL, 0);
        lst = GetObjectPtr(form, ctlID_FieldProperties_FieldsList);
        LstSetListChoices( lst, NULL, 0);

        for (i = 0; i < numItems; i++)
            DeallocateMemPtr( items[i]);
        if (items)
            DeallocateMemPtr( items);
        if (items2) {
            DeallocateMemPtr( items2);
        }

        Chooser_Close();

        if (source)
            CloseDataSource( source);

    if (closeAll)
        return false;
    else {
        FrmReturnToForm(0);
        return true;
    }
}

#ifdef CONFIG_SCRIPTS
static Boolean
Calculated_FieldPropertiesHandleEvent(EventPtr event)
{
    FormPtr form;
    FieldPtr fld;
    MemHandle h;
    UInt16 ObjectsTable[] =
    {
        9001,
        ctlID_FieldProperties_ScriptField,
        ctlID_FieldProperties_HelpScriptButton,
        0
    };

    switch (event->eType) {
    case frmOpenEvent:
        form = FrmGetActiveForm();
        if (!InitFieldPropertiesForm(form, ObjectsTable))
            return true;

        fld = GetObjectPtr(form, ctlID_FieldProperties_ScriptField);
        h = FldGetTextHandle( fld );
        FldSetTextHandle( fld, SchemaFields[FieldNumProperties].data.calc.scriptText );
        if ( h ) DeallocateMemHandle( h );

		lastItemSelected = 0; // we actually use this to count ')'s
        FrmDrawForm(form);
    return true;
    case frmCloseEvent:
        form = FrmGetActiveForm();
        fld = GetObjectPtr(form, ctlID_FieldProperties_ScriptField);
	    FldSetTextHandle(fld,NULL );
    return false;
    case ctlSelectEvent:
        switch (event->data.ctlSelect.controlID) {
       case ctlID_FieldProperties_HelpScriptButton:
            FrmHelp(stringID_Help_Script);
        return true;
        case ctlID_FieldProperties_OK:
		{
			ScriptContext sc;
			
			sc.src = NULL;
			sc.context = RECORD_FIELD_WRITING_CONTEXT;
			sc.flags = ScriptContext_DataAccess_DTC;
			sc.data_access.da.sdf = SchemaFields;
			sc.data_access.da.n = NumFields;
			sc.rec_field = FieldNumProperties;

            form = FrmGetActiveForm();
            fld = GetObjectPtr(form, ctlID_FieldProperties_ScriptField);
	        FldSetTextHandle(fld,NULL );

            if ( DesignViewDriver->class_ops.CompileScript( &sc, &SchemaFields[FieldNumProperties].data.calc, false ) ) {
                SchemaFields[FieldNumProperties].data.changed = true;
				ForceRebuild = true;
            } else {
                DeallocateMemHandle( SchemaFields[FieldNumProperties].data.calc.bytecode );
                SchemaFields[FieldNumProperties].data.calc.bytecode = NULL;
                FldSetTextHandle( fld, SchemaFields[FieldNumProperties].data.calc.scriptText );
				FldDrawField( fld );
                return true;
            }

            {
                ControlPtr ctl;
                ctl = GetObjectPtr(form, ctlID_FieldProperties_LockBox);
                SchemaFields[FieldNumProperties].lock = CtlGetValue(ctl);
            }
            FrmReturnToForm(0);
		}
        return true;
        case ctlID_FieldProperties_Cancel:
            form = FrmGetActiveForm();
            fld = GetObjectPtr(form, ctlID_FieldProperties_ScriptField);
	        FldSetTextHandle(fld,NULL );
            FrmReturnToForm(0);
        return true;
        default:
        }
    break;
	case keyDownEvent:

		/**
		 *  Active Parenthesis
		 */
		if ( event->data.keyDown.chr == ')' ) {
			Int16 q,p,l;
			UInt32 t;
			char *txt;

            form = FrmGetActiveForm();
            fld = GetObjectPtr(form, ctlID_FieldProperties_ScriptField);
			txt = MemHandleLock(FldGetTextHandle( fld ) );
			q = StrLen( txt ) - 1;
			p = 1;
			l = FldGetInsPtPosition( fld );
			while( q > 0) {
				if ( txt[q] == '(' ) p--;
				else if ( txt[q] == ')' ) p++;
				
				if ( p < 1 ) break;
				q--;
			}
			FldSetInsPtPosition( fld, q );
			t = TimGetTicks();
			q = 0;
			while( (TimGetTicks() - t) < SysTicksPerSecond() / 3 ) {
				p += q*11;
				q++;
			}
			MemPtrUnlock( txt );
			FldSetInsPtPosition( fld, l );
			
			
		}
		break;

    default:
    }
    return false;
}
#endif
